//
// starfld.c
//
// Library of routines to create animated starfields.
// For 3D displays only with origin in center of screen.
//
// richard.cutting@virgin.net 
// http://www.netyaroze-europe.com/~rcutting
//
// Please feel free to use this code as you want to for your own non
// commercial use but please leave my name in the headers if you do. If
// you do use it, extend it or change it then please let me know !
//
// Richard Cutting 14-May-1998
//

#include <libps.h>
#include <assert.h>

#include "starfld.h"

// We'll keep all our stars in here.

static StarType *GpStarHead = NULL;

// Lots of globals, mainly to avoid doing too many calculations per frame.

static int G_ScrollDir;		// direction scrolling
static int G_ChangeDirTime;	// time we'll change direction at
static int G_MaxX;		// maximum X 
static int G_MaxY;		// maximum Y
static int G_HalfX;		// G_MaxX / 2
static int G_HalfY;		// G_MaxY / 2
static int G_XLimit;		// X Limit
static int G_YLimit;		// Y Limit
static int G_Top;		// Top of screen
static int G_Bottom;		// Bottom of screen
static int G_Left;		// Left hand side of screen
static int G_Right;		// Right hand side of screen
static int G_AnimateFrame;	// Frame to animate on
static u_short G_ZPos;		// Priority in OT

//
// Free the memory used by the starfield.
//

void free_starfield( void )
{
	register StarType *p, *tmp;

	assert( GpStarHead != ( StarType * ) NULL );

	p = GpStarHead;

	while ( p )
	{
		tmp = p->next;
		free( p );
		p = tmp;
	}
	
	GpStarHead = ( StarType * ) NULL;
}		

//
// Initialise a starfield.
//
	
void initialise_starfield
(
	int num_stars,
	int num_planes,
	int scroll_dir,
	int change_time,
	int animate_every_n_frame,
	int ot_length
)
{	
	register int i, plane;

	StarType *sp, *p;

	assert( scroll_dir == STAR_NO_SCROLL || scroll_dir == STAR_SCROLL_LEFT || 
		scroll_dir == STAR_SCROLL_RIGHT || scroll_dir == STAR_SCROLL_UP ||
		scroll_dir == STAR_SCROLL_DOWN );
	
	G_ScrollDir = scroll_dir;
	G_ChangeDirTime = change_time;
	G_AnimateFrame = animate_every_n_frame;
	G_ZPos = ( u_short ) pow( ( double ) ot_length, 2 );
	G_MaxX = GsDISPENV.disp.w;
	G_MaxY = GsDISPENV.disp.h;

	// If we're being re-called then free up the existing field.

	if ( GpStarHead != ( StarType * ) NULL )
	{
		free_starfield();
	}

	// As we are using 3D then we need to adjust the x,y co-ordinates of the star
	// as our origin ( 0, 0 ) is at the centre of the screen. So if we are using
	// 320 * 240 then our co-ordinates actually run from -160,-120 to 160, 120. Note
	// that this doesn't take into account of your origin is not the centre of the
	// screen - you'll have to do that yourself !
	// ( Thanks to Robert Swan ! )

	G_HalfX = G_MaxX / 2;
	G_HalfY = G_MaxY / 2;
	G_XLimit = G_HalfX;
	G_YLimit = G_HalfY;
	G_Left = G_XLimit * -1;
	G_Right = G_XLimit;
	G_Top = G_YLimit * -1;
	G_Bottom = G_YLimit;

	p = GpStarHead;

	// Initialise each star in turn.

	for ( i = 0; i < num_stars; i++ )
	{
		sp = ( StarType * ) malloc( sizeof( StarType ) );
		assert( sp != ( StarType * ) NULL );

		sp->next = ( StarType * ) NULL;
		sp->box.x = ( rand()%G_MaxX ) - ( G_HalfX );
		sp->box.y = ( rand()%G_MaxY ) - ( G_HalfY );
		sp->box.w = sp->box.h = 1;

		// We are using more than one plane.

		if ( num_planes > 1 )
		{
			plane = ( rand()%num_planes ) + 1;
			sp->speed = ( num_planes - plane ) + 1;
			sp->box.r = sp->box.g = sp->box.b = 255 / plane;
		}
		else	// single plane
		{
			sp->speed = 1;
			sp->box.r = sp->box.g = sp->box.b = 128;
		}

		// Assign star into the singly linked list.

		if ( GpStarHead == ( StarType * ) NULL )
		{
			GpStarHead = p = sp;
		}
		else
		{
		     	p->next = sp;
		     	p = sp;
		}
	}
}

//
// Animate the starfield.
//

void animate_starfield( void )
{
	register StarType *s;

	register GsBOXF *box;

	int absX, absY, speed;

	static int AnimationTimer = 0;
	static int ChangeTimer = 0;

	assert( GpStarHead != ( StarType * ) NULL );

	AnimationTimer++;
	ChangeTimer++;

	// If we've got nothing to do then get out quickly.

	if ( G_ScrollDir == STAR_NO_SCROLL || 
	     AnimationTimer != G_AnimateFrame )
	{
		return;
	}

	// Time to swap scrolling direction ?

	if ( G_ChangeDirTime == ChangeTimer )
	{
		switch( G_ScrollDir )
		{
			case STAR_SCROLL_RIGHT:
				G_ScrollDir = STAR_SCROLL_LEFT;
				break;	
			case STAR_SCROLL_LEFT:
				G_ScrollDir = STAR_SCROLL_RIGHT;
				break;
			case STAR_SCROLL_UP:
				G_ScrollDir = STAR_SCROLL_DOWN;
				break;
			case STAR_SCROLL_DOWN:
				G_ScrollDir = STAR_SCROLL_UP;
				break;
		};

		ChangeTimer = 0;
 	}

	// Move each star in our linked list.

	s = GpStarHead;

	while( s )
	{
		// Assign for speed...

		box = &( s->box );
		speed = s->speed;

		// Move the star by adjusting x,y co-ordinates accordingly

		if ( G_ScrollDir == STAR_SCROLL_RIGHT )
			box->x += speed;
		else if ( G_ScrollDir == STAR_SCROLL_LEFT )
			box->x -= speed;
		else if ( G_ScrollDir == STAR_SCROLL_UP )
			box->y -= speed;
		else if ( G_ScrollDir == STAR_SCROLL_DOWN )
			box->y += speed;

		// If it has moved off the screen then reposition on the
		// correct edge.

		absX = abs( box->x );
		absY = abs( box->y );

		if ( absY > G_YLimit )
		{
			if ( G_ScrollDir == STAR_SCROLL_UP )
				box->y = G_Bottom;
			else if ( G_ScrollDir == STAR_SCROLL_DOWN )
				box->y = G_Top;
		}

		if ( absX > G_XLimit  )
		{
			if ( G_ScrollDir == STAR_SCROLL_RIGHT )
				box->x = G_Left;
			else if ( G_ScrollDir == STAR_SCROLL_LEFT )
				box->x = G_Right;
		}

		s = s->next;
	}

	AnimationTimer = 0;	
}

//
// Load starfield into OT.
//

void load_starfield_into_OT( GsOT *otp )
{
	register StarType *sp;
	
	sp = GpStarHead;

	while( sp )
	{
		GsSortBoxFill( &( sp->box ), otp, G_ZPos );
		sp = sp->next;
	}
}

